<?php


namespace Schaffen\ProductionOrder\ProductionOrder\Managers;


use App\RestApi\ProductionOrder\Resources\ProductionOrderResource;
use Carbon\Carbon;
use Exception;
use Freelabois\LaravelQuickstart\Extendables\ManipulationManager;
use Illuminate\Support\Facades\DB;
use Schaffen\Core\EnvironmentConfiguration\Repositories\ParameterRepository;
use Schaffen\Core\Movement\Managers\ItemMovementBatchManager;
use Schaffen\Core\Product\Models\Batch;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\CompositeProductRepository;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\ProductionOrderItemRepository;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\ProductionOrderRepository;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\StatusRepository;
use Schaffen\WarehouseInventory\WarehouseInventory\Services\SnapshotService;
use Schaffen\ProductionOrder\ProductionOrder\Repositories\CompositeProductHistoricRepository;
use Schaffen\Core\Product\Repositories\ProductVarietyPackRepository;


class ProductionOrderManager extends ManipulationManager
{
    protected CompositeProductRepository $composite_product_repository;
    protected ProductionOrderItemManager $production_order_item_manager;
    protected ProductionOrderItemRepository $production_order_item_repository;
    protected ProductionOrderStatusManager $production_order_status_manager;
    protected StatusRepository $status_repository;
    protected SnapshotService $snapshot_service;
    protected ParameterRepository $parameter_repository;
    protected CompositeProductHistoricRepository $composite_product_historic_repository;
    protected ItemMovementBatchManager $item_movement_batch_manager;

    public function __construct(ProductionOrderRepository $repository)
    {
        parent::__construct($repository);

        $this->composite_product_repository = app()->make(CompositeProductRepository::class);
        $this->production_order_item_manager = app()->make(ProductionOrderItemManager::class);
        $this->production_order_status_manager = app()->make(ProductionOrderStatusManager::class);

        $this->status_repository = app()->make(StatusRepository::class);
        $this->snapshot_service = app()->make(SnapshotService::class);
        $this->parameter_repository = app()->make(ParameterRepository::class);


        $this->production_order_item_repository = app()->make(ProductionOrderItemRepository::class);

        $this->composite_product_historic_repository = app()->make(CompositeProductHistoricRepository::class);

        $this->product_variety_pack_repository = app()->make(ProductVarietyPackRepository::class);
        $this->item_movement_batch_manager = app()->make(ItemMovementBatchManager::class);
    }

    public function index($filter)
    {
        return $this->repository->list($filter, [
            'finished_product',
            'finished_product.measurement_unit',
            'finished_product.compositionQuantity',
            'status',
        ]);
    }

    public function storeOrUpdate($values, $id = null, array $relations = [])
    {
        //gravar ordem de producao com status aguardando producao

        return DB::transaction(function () use ($values, $id, $relations) {

            $finished_product = $this->product_variety_pack_repository->first([
                'pve_codigo' => $values['op_fk_pve_final']
            ]);

            if (!$finished_product) {
                throw new \Exception("O produto  selecionado não existe.");
            }

            $composition_products = $this->composite_product_repository->findMany(
                [
                    "pcp_fk_pve_codigo_acabado" => $values['op_fk_pve_final']
                ]
            );

            if ($composition_products->isEmpty()) {
                throw new \Exception("O produto {$finished_product->pve_codigo} - {$finished_product->product_complete_name} não é de composição. "
                    . "Por favor, informe um produto do tipo composição para cadastrar uma ordem de produção.");
            }

            $values['op_quantidade_produzida'] = $values['op_quantidade'];
            $production_order = $this->repository->storeOrUpdate($values)->load(['finished_product']);

            $recipes = floatval($values['op_qtd_receitas']);

            $composition_products = $composition_products->map(function ($product) use ($production_order, $recipes) {
                $product_variety_pack = $this->product_variety_pack_repository->firstOrFail([
                    'pve_codigo' => $product->pcp_fk_pve_codigo_materia_prima
                ], ['product_variety', 'product_variety.product']);

                $warehouse_inventory = $this->snapshot_service->getInventoryFromBranch($this->getBranchId());

                $total_value = 0;
                if (boolval($product_variety_pack->product_variety->product->pro_servico)) {
                    $total_value = $product_variety_pack->product_variety->prv_valor_venda;
                } else {
                    try {
                        $snapshot = $this->snapshot_service->validateProductInventory(
                            $warehouse_inventory->loe_codigo,
                            $product->pcp_fk_pve_codigo_materia_prima,
                            0
                        );

                        $total_value = $product->pcp_quantidade * $recipes * $snapshot->sna_valor_medio;
                    } catch (Exception $e) {
                        throw new \Exception("A matéria-prima selecionada ({$product->pcp_fk_pve_codigo_materia_prima} - {$product_variety_pack->product_complete_name}) não possui custo médio para ser utilizado no cálculo do custo de produção.");
                    }
                }

                return [
                    "opi_fk_pve_codigo" => $product->pcp_fk_pve_codigo_materia_prima,
                    "opi_fk_op_codigo" => $production_order->op_codigo,
                    "opi_quantidade_inicial" => $product->pcp_quantidade * $recipes,
                    "opi_quantidade_final" => $product->pcp_quantidade,
                    "opi_controla_estoque" => $product->pcp_controla_estoque,
                    "opi_valor_inicial" => $total_value,
                    "opi_valor_final" => $total_value
                ];
            });

            $production_order_items = [];
            foreach ($composition_products as $product) {
                $production_order_items[] = $this->production_order_item_repository->storeOrUpdate($product)->load(['feedstock']);
            }

            $this->composite_product_historic_repository->storeOrUpdateOP(
                $production_order,
                $production_order_items,
                CompositeProductHistoricRepository::STORE_OP
            );

            $status = $this->status_repository->firstOrFail();

            $production_order_status = [
                "ops_fk_op_codigo" => $production_order->op_codigo,
                "ops_fk_status" => $status->st_codigo
            ];

            $this->production_order_status_manager->storeOrUpdate($production_order_status);

            return new ProductionOrderResource($production_order->load('status_po', 'finished_product'));
        });
    }

    public function update($values, $id)
    {
        $production_order = $this->repository->firstOrFail([
            'op_codigo' => $id
        ], [
            'status',
        ]);

        if (!$production_order->status[0]->editable()) {
            throw new \Exception("A ordem de produção não pode ser alterada no status atual");
        }

        return DB::transaction(function () use ($production_order, $values) {
            $production_order->fill($values);
            $production_order->save();

            $composition_products = $this->composite_product_repository->findMany(
                [
                    "pcp_fk_pve_codigo_acabado" => $production_order->op_fk_pve_final,
                ]
            );

            $composition_products = $composition_products->map(function ($product) use ($production_order) {
                $production_order_item = $this->production_order_item_repository->first([
                    'opi_fk_op_codigo' => $production_order->op_codigo,
                    'opi_fk_pve_codigo' => $product->pcp_fk_pve_codigo_materia_prima,
                ]);

                if (!$production_order_item) {
                    throw new \Exception("A produção da máteria-prima " . $product->pcp_fk_pve_codigo_materia_prima . " não foi encontrada");
                }

                $product_variety_pack = $this->product_variety_pack_repository->firstOrFail([
                    'pve_codigo' => $product->pcp_fk_pve_codigo_materia_prima
                ], ['product_variety', 'product_variety.product']);

                $warehouse_inventory = $this->snapshot_service->getInventoryFromBranch($this->getBranchId());

                $total_value = 0;

                if (boolval($product_variety_pack->product_variety->product->pro_servico)) {
                    $total_value = $product_variety_pack->product_variety->prv_valor_venda;
                } else {
                    try {
                        $snapshot = $this->snapshot_service->validateProductInventory(
                            $warehouse_inventory->loe_codigo,
                            $product->pcp_fk_pve_codigo_materia_prima,
                            0
                        );

                        $total_value = $product->pcp_quantidade * floatval($production_order->op_qtd_receitas) * $snapshot->sna_valor_medio;
                    } catch (Exception $e) {
                        throw new \Exception("A matéria-prima selecionada ({$product->pcp_fk_pve_codigo_materia_prima} - {$product_variety_pack->product_complete_name}) não possui custo médio para ser utilizado no cálculo do custo de produção.");
                    }
                }

                return [
                    "opi_codigo" => $production_order_item->opi_codigo,
                    "opi_fk_pve_codigo" => $product->pcp_fk_pve_codigo_materia_prima,
                    "opi_fk_op_codigo" => $production_order->op_codigo,
                    "opi_quantidade_inicial" => $product->pcp_quantidade * floatval($production_order->op_qtd_receitas),
                    "opi_quantidade_final" => $product->pcp_quantidade * floatval($production_order->op_qtd_receitas),
                    "opi_controla_estoque" => $product->pcp_controla_estoque,
                    "opi_valor_inicial" => $total_value,
                    "opi_valor_final" => $total_value
                ];
            });


            $production_order_items = [];

            foreach ($composition_products as $product) {
                $production_order_items[] = $this->production_order_item_repository->storeOrUpdate(
                    $product,
                    $product['opi_codigo']
                )->load(['feedstock']);
            }

            $this->composite_product_historic_repository->storeOrUpdateOP(
                $production_order,
                $production_order_items,
                CompositeProductHistoricRepository::UPDATE_OP
            );

            return $production_order;
        });
    }

    public function show($id)
    {
        return $this->repository->firstOrFail([
            'op_codigo' => $id
        ], [
            'finished_product',
            'finished_product.measurement_unit',
            'finished_product.compositionQuantity',
            'items',
            'items.feedstock',
            'items.feedstock.product_variety',
            'items.feedstock.measurement_unit'
        ]);
    }

    public function showHistoric($id)
    {
        return $this->composite_product_historic_repository
            ->setOrder('created_at', 'desc')
            ->findMany([
                'pch_fk_op_codigo' => $id,
                'operation_in' => [
                    CompositeProductHistoricRepository::STORE_OP,
                    CompositeProductHistoricRepository::UPDATE_OP,
                    CompositeProductHistoricRepository::UPDATE_OP_STATUS,
                    CompositeProductHistoricRepository::FINALIZATION_OP,
                    CompositeProductHistoricRepository::DESTROY_OP
                ]
            ]);
    }

    public function changeProductionOrderStatus(array $values, int $id)
    {

        return DB::transaction(function () use ($values, $id) {
            $production_order = $this->repository->find($id, ['status_po', 'status', 'finished_product']);

            $status = $this->status_repository->find($values['st_codigo']);

            if ($status->st_codigo == $production_order->status_po->ops_fk_status) {
                throw new \Exception("Essa ordem já se encontra no status '$status->st_descricao'");
            }

            if ($production_order->status_po->status->isFinal()) {
                throw new \Exception("Essa ordem de produção já se encontra finalizada!");
            }


            $warehouse_inventory_origin = null;
            $warehouse_inventory_destiny = null;
            $snap_op_mode = null;

            extract($this->status_repository->getWarehouseFlow($production_order->status_po->status, $status));

            if ($warehouse_inventory_origin) {

                $items = $this->production_order_item_repository->findMany(["opi_fk_op_codigo" => $id], [
                    "feedstock",
                    "feedstock.product_variety",
                    "feedstock.product_variety.product"
                ]);

                $op_value = 0;
                foreach ($items as $item) {

                    if (!boolval($item->opi_controla_estoque) || boolval($item->feedstock->product_variety->product->pro_servico)) {
                        continue;
                    }

                    $quantity = null;

                    if (!$this->parameter_repository->find(23)->par_valor) {
                        $quantity = $item->opi_quantidade_inicial;
                    }

                    try {
                        $snapshot = $this->snapshot_service->validateProductInventory(
                            $warehouse_inventory_origin,
                            $item->opi_fk_pve_codigo,
                            $quantity,
                        );
                    } catch (\Exception $e) {
                        throw new \Exception("Produto nº {$item->opi_fk_pve_codigo} não possui estoque para a" .
                            " quantidade {$item->opi_quantidade_inicial}.");
                    }

                    if (!boolval($status->isFinal())) {

                        $this->snapshot_service->createAccountInventory(
                            $id,
                            $item->opi_quantidade_inicial,
                            $item->opi_fk_pve_codigo,
                            $snapshot->sna_valor_medio,
                            $warehouse_inventory_origin,
                            0,
                            SnapshotService::ORDEM_PRODUCAO
                        );

                        $this->snapshot_service->createAccountInventoryProductionOrder(
                            $id,
                            $item->opi_quantidade_inicial,
                            $item->opi_fk_pve_codigo,
                            $snapshot->sna_valor_medio,
                            $warehouse_inventory_destiny,
                            1,
                            SnapshotService::ORDEM_PRODUCAO,
                            $snap_op_mode
                        );

                        $item->opi_fk_loe_codigo_origem = $warehouse_inventory_origin;
                        $item->opi_fk_loe_codigo_destino = $warehouse_inventory_destiny;

                        $item->save();
                    } else {
                        $this->snapshot_service->createAccountInventory(
                            $id,
                            $item->opi_quantidade_inicial,
                            $item->opi_fk_pve_codigo,
                            $snapshot->sna_valor_medio,
                            $warehouse_inventory_origin,
                            0,
                            SnapshotService::ORDEM_PRODUCAO,
                        );

                        $op_value += $snapshot->sna_valor_medio * $item->opi_quantidade_inicial;
                    }
                }

                if ($status->isFinal()) {
                    if (!isset($values['op_quantidade_produzida'])) {
                        throw new \Exception("Informe a quantidade produzida para finalizar a ordem de produção!");
                    }

                    if (!empty($values['op_observacao'])) {
                        $production_order->op_observacao = $values['op_observacao'];
                    }

                    $production_order->op_quantidade_produzida = $values['op_quantidade_produzida'];
                    $production_order->save();

                    $warehouse_inventory_destiny = $this->snapshot_service->getInventoryFromBranch($this->getBranchId());

                    if ($production_order->finished_product->pve_controla_lote) {
                        $this->item_movement_batch_manager->storeOrUpdate([
                            "mil_operacao" => "ORDEM_PRODUCAO",
                            "mil_fk_mov_codigo" => $production_order->op_codigo,
                            "mil_fk_lot_codigo" => $values['lot_codigo'],
                            "mil_fk_pve_codigo" => $production_order->op_fk_pve_final,
                            "mil_fk_loe_codigo" => $warehouse_inventory_destiny->loe_codigo,
                            "mil_quantidade" => $production_order->op_quantidade,
                            "mil_data_cancelamento" => null,
                            "mil_sentido" => 1,
                        ]);
                    }

                    $this->snapshot_service->createAccountInventoryProductionOrder(
                        $id,
                        $values['op_quantidade_produzida'],
                        $production_order->op_fk_pve_final,
                        $op_value / $values['op_quantidade_produzida'],
                        $warehouse_inventory_destiny->loe_codigo,
                        1,
                        SnapshotService::ORDEM_PRODUCAO,
                        $snap_op_mode
                    );
                }
            }

            $this->production_order_status_manager->storeOrUpdate(
                ["ops_fk_status" => $status->st_codigo],
                $production_order->status_po->ops_codigo
            );


            $this->composite_product_historic_repository->changeStatusOP(
                $production_order,
                [
                    'old_status' => $production_order->status_po->status,
                    'new_status' => $status
                ]
            );


            return $production_order->load('status_po');
        });
    }

    public function destroy($id)
    {
        $production_order = $this->repository->find($id, ['status_po', 'status', 'finished_product', 'items']);
        if (!$production_order->status_po->status->editable()) {
            throw new \Exception('A ordem de produção selecionada se encontra em um status que não permite exclusão!');
        }

        return DB::transaction(function () use ($production_order) {

            $items = [];
            foreach ($production_order->items as $item) {
                $this->production_order_item_repository->destroy($item->opi_codigo);
                $items[] = $item;
            }

            $production_order->status_po->destroy($production_order->status_po->ops_codigo);
            $this->repository->destroy($production_order->op_codigo);

            $this->composite_product_historic_repository->storeOrUpdateOP(
                $production_order,
                $items,
                CompositeProductHistoricRepository::DESTROY_OP
            );

            return $production_order;
        });
    }

    private function getBranchId()
    {
        return 1;
    }

    public function batchList($id)
    {
        $data = Batch::whereHas(
            'product_batch',
            function ($query) use ($id) {
                $query->where('lopv_fk_pve_codigo', $id);
            }
        )
            ->where('lot_status', 1)
            ->where('lot_validade', '>', Carbon::now()->format('Y-m-d'))
            ->get();

        return $data;
    }
}
